import requests
import time
import os
import json
from uuid import uuid4

endpoint = 'https://api.booste.io/'
# Endpoint override for development
if 'BOOSTE_URL' in os.environ:
    print("Dev Mode")
    if os.environ['BOOSTE_URL'] == 'local':
        endpoint = 'http://localhost/'
    else:
        endpoint = os.environ['BOOSTE_URL']
    print("Hitting endpoint:", endpoint)


# identify machine for blind use
cache_path = os.path.abspath(os.path.join(os.path.expanduser('~'),".booste","cache.json"))
if os.path.exists(cache_path):
    with open(cache_path, "r") as file:
        cache = json.load(file)
else:
    cache = {}
    cache['machine_id'] = str(uuid4())
    os.makedirs(os.path.join(os.path.expanduser('~'), ".booste"), exist_ok=True)
    with open(cache_path, "w+") as file:
        json.dump(cache, file)


client_error = {
    "OOB" : "Client error: {}={} is out of bounds.\n\tmin = {}\n\tmax = {}"
}



# THE MAIN FUNCTIONS
# ___________________________________

# The syncronous end-to-end caller. Takes in user prams, returns outstring, throws error if server is shit
def gpt2_sync_main(api_key, model_size, in_string, length, temperature, window_max):
    sync_mode = "synchronous"
    validate_input(temperature, window_max)
    task_id = call_start_api(api_key, sync_mode, model_size, in_string, length, temperature, window_max)
    interval, initial_wait = choose_delay_params(model_size, length, window_max)
    time.sleep(initial_wait)
    while True:
        dict_out = call_check_api(api_key, sync_mode, task_id)
        if dict_out['Status'] == "Finished":
            return dict_out["Output"]
        if dict_out["Status"] == "Failed":
            raise Exception("Server error: Booste inference job returned status 'Failed'")
        time.sleep(interval)

def gpt2_async_start_main(api_key, model_size, in_string, length, temperature, window_max):
    sync_mode = "asynchronous"
    validate_input(temperature, window_max)
    task_id = call_start_api(api_key, sync_mode, model_size, in_string, length, temperature, window_max)
    return task_id

def gpt2_async_check_main(api_key, task_id):
    sync_mode = "asynchronous"
    dict_out = call_check_api(api_key, sync_mode, task_id)
    return dict_out




# THE API CALLING FUNCTIONS
# ________________________


# The bare async starter. Used by both gpt2_sync_main (automated) and async (client called)
# Takes in start params, returns task ID
def call_start_api(api_key, sync_mode, model_size, in_string, length, temperature, window_max):
    global endpoint
    route_start = 'inference/pretrained/gpt2/async/start'
    url_start = endpoint + route_start

    global cache
    # sequence = []
    payload = {
        "string" : in_string,
        "length" : str(length),
        "temperature" : str(temperature),
        "machineID" : cache['machine_id'],
        "apiKey" : api_key,
        "modelSize" : model_size,
        "windowMax" : window_max, 
        "syncMode": sync_mode
    }
    response = requests.post(url_start, json=payload)
    if response.status_code != 200:
        raise Exception("Server error: Booste inference server returned status code {}\n{}".format(
            response.status_code, response.json()['message']))
    
    try:
        out = response.json()
        task_id = out['TaskID']
        return task_id
    except:
        raise Exception("Server error: Failed to return TaskID")

# The bare async checker. Used by both gpt2_sync_main (automated) and async (client called)
# Takes in task ID, returns reformatted dict_out with Status and Output
def call_check_api(api_key, sync_mode, task_id):
    global endpoint
    route_check = 'inference/pretrained/gpt2/async/check/v2'
    url_check = endpoint + route_check

    # Poll server for completed task
    payload = {"TaskID": task_id, "apiKey": api_key, "syncMode": sync_mode}
    response = requests.post(url_check, json=payload)
    if response.status_code != 200:
        raise Exception("Server error: Booste inference server returned status code {}\n{}".format(
            response.status_code, response.json()['message']))
    out = response.json()
    return out




# THE MISC FUNCTIONS
# ___________________________________

# The function to raise exceptions if parameters are not valid
def validate_input(temperature, window_max):
    # Make sure request is valid
    global client_error
    if temperature < 0.1 or temperature > 1:
        raise Exception(client_error['OOB'].format("temperature", temperature, "0.1", "1"))
    if window_max < 1 or window_max > 1023:
        raise Exception(client_error['OOB'].format("window_max", window_max,   "1", "1023"))

# The function to determine the frequency that the syncronous functions polls for an async response
def choose_delay_params(model_size, length, window_max): # window max not yet accounted for
    # Choose a delay approprate for the call
    if model_size == 'gpt2':
        interval = length * 0.1
        initial_wait = length * .2
    elif model_size == 'gpt2-xl':
        interval = length * 0.2
        initial_wait = length * .4
    else:
        interval = 3
        initial_wait = length * .3
    # Correct for small calls so it's not rediculous
    if interval < 3:
        interval = 3
    return interval, initial_wait


import os
from uuid import uuid4
import json
import time
from .gpt2_utils import gpt2_sync_main, gpt2_async_start_main, gpt2_async_check_main
from .generics import run_main, start_main, check_main
from .clip_utils import clip_main, async_clip_main, clip_image_main, clip_text_main
import asyncio
import sys

# Generics
def run(api_key, model_key, model_parameters):
    out = run_main(
        api_key = api_key, 
        model_key = model_key, 
        model_parameters = model_parameters
    )
    return out

def start(api_key, model_key, model_parameters):
    out = start_main(
        api_key = api_key, 
        model_key = model_key, 
        model_parameters = model_parameters
    )
    return out
    
def check(api_key, task_id):
    out_dict = check_main(
        api_key = api_key,
        task_id = task_id
    )
    return out_dict


# GPT2 small
def gpt2(api_key, in_string, length = 5, temperature = 0.8, window_max = 100):
    out_list = gpt2_sync_main(
        api_key = api_key, 
        model_size = "gpt2", 
        in_string = in_string, 
        length = length, 
        temperature = temperature,  
        window_max = window_max)
    return out_list

def gpt2_async_start(api_key, in_string, length = 5, temperature = 0.8, window_max = 100):
    task_id = gpt2_async_start_main(
        api_key = api_key, 
        model_size = "gpt2", 
        in_string = in_string, 
        length = length, 
        temperature = temperature,  
        window_max = window_max)
    return task_id

def gpt2_async_check(api_key, task_id):
    dict_out = gpt2_async_check_main(api_key, task_id)
    return dict_out

# GPT2 XL
def gpt2_xl(api_key, in_string, length = 5, temperature = 0.8, window_max = 100):
    out_list = gpt2_sync_main(
        api_key = api_key, 
        model_size = "gpt2-xl", 
        in_string = in_string, 
        length = length, 
        temperature = temperature,  
        window_max = window_max)
    return out_list

def gpt2_xl_async_start(api_key, in_string, length = 5, temperature = 0.8, window_max = 100):
    task_id = gpt2_async_start_main(
        api_key = api_key, 
        model_size = "gpt2-xl", 
        in_string = in_string, 
        length = length, 
        temperature = temperature,  
        window_max = window_max)
    return task_id

def gpt2_xl_async_check(api_key, task_id):
    dict_out = gpt2_async_check_main(api_key, task_id)
    return dict_out

def clip(api_key, prompts, images, pretty_print=False):
    try:
        # Run it async
        dict_out = asyncio.run(async_clip_main(api_key, prompts, images, pretty_print))
        return dict_out

    except:
        # Run it sync if async failed
        dict_out = clip_main(api_key,prompts, images, pretty_print)
        return dict_out
        

# def clip_image(images):
#     dict_out = clip_image_main(images)

# def clip_text(prompts):
#     dict_out = clip_text_main(prompts)

# def clip_fast(encoded_prompts, encoded_images, forward=True):
#     dict_out = clip_fast_main(encoded_prompts, encoded_images, forward)


import os
import asyncio
import requests
import numpy as np
import json
from PIL import Image
import base64
from io import BytesIO
import re
import math

url_regex = re.compile(
        r'^(?:http|ftp)s?://' # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
        r'localhost|' #localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
        r'(?::\d+)?' # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

endpoint = 'https://7rq1vzhvxj.execute-api.us-west-1.amazonaws.com/Prod/infer/'
# Endpoint override for development
devmode = False
if 'BoosteURL' in os.environ:
    print("Dev Mode")
    devmode = True
    if os.environ['BoosteURL'] == 'local':
        endpoint = 'http://localhost:8080/2015-03-31/functions/function/invocations'
    else:
        endpoint = os.environ['BoosteURL']
    print("Hitting endpoint:", endpoint)


async def async_api_caller(api_key, prompt, image):
    global endpoint
    url = endpoint
    # print(url)

    # defaults
    is_path = False
    is_url = False

    if re.match(url_regex, image) is not None:
        is_url = True
        image_send = image

    if os.path.exists(image):
        is_path = True

        if image[-3:] == "jpg" or image[-4:] == "jpeg":
            image_pil = Image.open(image)
            image_file = BytesIO()
            image_pil.save(image_file, format="JPEG")
            image_bytes = image_file.getvalue()  # im_bytes: image in binary format.
            image_send = base64.b64encode(image_bytes).decode('utf-8')
        elif image[-3:] == "png":
            image_pil = Image.open(image)
            image_file = BytesIO()
            image_pil.save(image_file, format="PNG")
            image_bytes = image_file.getvalue()  # im_bytes: image in binary format.
            image_send = base64.b64encode(image_bytes).decode('utf-8')
        else:
            print("Warning, image failed: {}\nImage must be .jpg or .png.\n".format(image))
            return None

    if is_path == False and is_url == False:
        print("Warning: image failed: {}\nImage must be valid a URL or a path to local image file\n\texample:  https://google.com\n\texample:  ./my-image.jpg\n\texample:  /home/user/my-image.png\n".format(image))
        return None

    # sequence = []
    payload = {
        "apiKey" : api_key,
        "prompt" : prompt,
        "image" : image_send,
        "isUrl" : is_url
    }

    response = requests.post(url, json=payload)
    try:
        if devmode: # then it returns as full json blob
            out = response.json()
            code = out['statusCode']
            body = json.loads(out['body'])
        else:
            # then api gateway converts status code in blob to status code, and body in json to json
            code = response.status_code
            body = response.json()

    except Exception as e:
        print("Warning, server failed to process one request:\n\tprompt:  {}\n\timage:  {}\n".format(prompt, image))
        return None
        
    if code == 200:
        return body
    else:
        raise Exception("Server returned error code {}\n{}".format(code,body))



def api_caller(api_key, prompt, image):
    global endpoint
    url = endpoint
    # print(url)

    # defaults
    is_path = False
    is_url = False

    if re.match(url_regex, image) is not None:
        is_url = True
        image_send = image

    if os.path.exists(image):
        is_path = True

        if image[-3:] == "jpg" or image[-4:] == "jpeg":
            image_pil = Image.open(image)
            image_file = BytesIO()
            image_pil.save(image_file, format="JPEG")
            image_bytes = image_file.getvalue()  # im_bytes: image in binary format.
            image_send = base64.b64encode(image_bytes).decode('utf-8')
        elif image[-3:] == "png":
            image_pil = Image.open(image)
            image_file = BytesIO()
            image_pil.save(image_file, format="PNG")
            image_bytes = image_file.getvalue()  # im_bytes: image in binary format.
            image_send = base64.b64encode(image_bytes).decode('utf-8')
        else:
            print("Warning, image failed: {}\nImage must be .jpg or .png.\n".format(image))
            return None

    if is_path == False and is_url == False:
        print("Warning: image failed: {}\nImage must be valid a URL or a path to local image file\n\texample:  https://google.com\n\texample:  ./my-image.jpg\n\texample:  /home/user/my-image.png\n".format(image))
        return None

    # sequence = []
    payload = {
        "apiKey" : api_key,
        "prompt" : prompt,
        "image" : image_send,
        "isUrl" : is_url
    }

    response = requests.post(url, json=payload)
    try:
        if devmode: # then it returns as full json blob
            out = response.json()
            code = out['statusCode']
            body = json.loads(out['body'])
        else:
            # then api gateway converts status code in blob to status code, and body in json to json
            code = response.status_code
            body = response.json()

    except Exception as e:
        print("Warning, server failed to process one request:\n\tprompt:  {}\n\timage:  {}\n".format(prompt, image))
        return None
        
    if code == 200:
        return body
    else:
        raise Exception("Server returned error code {}\n{}".format(code,body))


def softmax_caller(similarities, api_key):
    global endpoint
    url = endpoint

    # sequence = []
    payload = {
        "similarities" : similarities,
        "softmax" : True,
        "apiKey" : api_key
    }

    response = requests.post(url, json=payload)

    try:
        if devmode: # then it returns as full json blob
            out = response.json()
            code = out['statusCode']
            body = json.loads(out['body'])
        else:
            # then api gateway converts status code in blob to status code, and body in json to json
            code = response.status_code
            body = response.json()

    except Exception as e:
        print("Warning, server failed to run the probability step.")
        return None
        
    if code == 200:
        return body
    else:
        raise Exception("Server returned error code {}\n{}".format(code,body))


# # correct solution:
# def softmax(x):
#     """Compute softmax values for each sets of scores in x."""
#     e_x = np.exp(x - np.max(x))
    # return e_x / e_x.sum(axis=0) # only difference

def clip_main(api_key, prompts, images, pretty_print):
    if type(prompts) != type([]):
        raise Exception("Error: prompts not of type: list")
    
    if type(images) != type([]):
        raise Exception("Error: images not of type: list")

    if prompts == []:
        raise Exception("Error: prompts cannot be length: 0")
    else:
        for item in prompts:
            if type(item) != type(""):
                raise Exception("Error: all prompts must be type: string")

    if images == []:
        raise Exception("Error: images cannot be length: 0")
    else:
        for item in images:
            if type(item) != type(""):
                raise Exception("Error: all images must be type: string")


    outs = {}
    out_logits = np.zeros((len(prompts), len(images)))
    for i, prompt in enumerate(prompts):
        outs[prompt] = {}
        for j, image in enumerate(images):
            out = api_caller(api_key, prompt, image)
            outs[prompt][image] = out
            if out != None:
                out_logits[i,j] = out['similarity']

    out_probs = softmax_caller(out_logits.tolist(), api_key)
    if out_probs != None:
        for i, prompt in enumerate(prompts):
            for j, image in enumerate(images):
                if outs[prompt][image] != None:
                    outs[prompt][image]["probabilityRelativeToPrompts"] = out_probs["relativeToPrompts"][i][j]
                    outs[prompt][image]["probabilityRelativeToImages"] = out_probs["relativeToImages"][i][j]
    
    if pretty_print:
        pretty_print_output(prompts, images, outs)

    return outs

async def async_clip_main(api_key, prompts, images, pretty_print):

    if type(prompts) != type([]):
        raise Exception("Error: prompts not of type: list")
    
    if type(images) != type([]):
        raise Exception("Error: images not of type: list")

    if prompts == []:
        raise Exception("Error: prompts cannot be length: 0")
    else:
        for item in prompts:
            if type(item) != type(""):
                raise Exception("Error: all prompts must be type: string")

    if images == []:
        raise Exception("Error: images cannot be length: 0")
    else:
        for item in images:
            if type(item) != type(""):
                raise Exception("Error: all images must be type: string")


    tasks = {}
    for prompt in prompts:
        tasks[prompt] = {}
        for image in images:
            tasks[prompt][image] = asyncio.create_task(async_api_caller(api_key, prompt, image))

    outs = {}
    out_logits = np.zeros((len(prompts), len(images)))
    for i, prompt in enumerate(prompts):
        outs[prompt] = {}
        for j, image in enumerate(images):
            out = await tasks[prompt][image]
            outs[prompt][image] = out
            if out != None:
                out_logits[i,j] = out['similarity']
    # print(out_logits.tolist())

    out_probs = softmax_caller(out_logits.tolist(), api_key)
    if out_probs != None:
        for i, prompt in enumerate(prompts):
            for j, image in enumerate(images):
                if outs[prompt][image] != None:
                    outs[prompt][image]["probabilityRelativeToPrompts"] = out_probs["relativeToPrompts"][i][j]
                    outs[prompt][image]["probabilityRelativeToImages"] = out_probs["relativeToImages"][i][j]
    
    if pretty_print:
        pretty_print_output(prompts, images, outs)

    return outs

def pretty_print_output(prompts, images, outs):

    print("\n---------\nSimilarity scores\n---------\n")
    for prompt in prompts:
        for image in images:
            if outs[prompt][image] != None:
                print("\n\tPrompt:\t\t", prompt, "\n\tImage URL:\t", image, "\n\tSimilarity:\t", outs[prompt][image]["similarity"])
    print("\n")

    print("\n---------\nPrompt probabilities for each image:\n---------\n")
    for image in images:
        print("\tImage:", image)
        for prompt in prompts:
            if outs[prompt][image] != None:
                print("\n\t\tPrompt:", prompt, "\n\t\tProb:", outs[prompt][image]["probabilityRelativeToPrompts"])
            else:
                print("\n\t\tPrompt:", prompt, "\n\t\tFailed")
        print()
    print("\n")

    print("\n---------\nImage probabilities for each prompt:\n---------\n")
    for prompt in prompts:
        print("\tPrompt:", prompt)
        for image in images:
            if outs[prompt][image] != None:
                print("\n\t\tImage:", image, "\n\t\tProb:", outs[prompt][image]["probabilityRelativeToImages"])
            else:
                print("\n\t\tImage:", image, "\n\t\tFailed")
        print()
    print("\n")

def clip_image_main(images):
    return dict_out

def clip_text_main(prompts):
    return dict_out


import requests
import time
import os
import json
from uuid import uuid4

endpoint = 'https://api.booste.io/'
# Endpoint override for development
if 'BOOSTE_URL' in os.environ:
    print("Dev Mode")
    if os.environ['BOOSTE_URL'] == 'local':
        endpoint = 'http://localhost/'
    else:
        endpoint = os.environ['BOOSTE_URL']
    print("Hitting endpoint:", endpoint)


# identify machine for blind use
cache_path = os.path.abspath(os.path.join(os.path.expanduser('~'),".booste","cache.json"))
if os.path.exists(cache_path):
    with open(cache_path, "r") as file:
        cache = json.load(file)
else:
    cache = {}
    cache['machine_id'] = str(uuid4())
    os.makedirs(os.path.join(os.path.expanduser('~'), ".booste"), exist_ok=True)
    with open(cache_path, "w+") as file:
        json.dump(cache, file)


client_error = {
    "OOB" : "Client error: {}={} is out of bounds.\n\tmin = {}\n\tmax = {}"
}

# THE MAIN FUNCTIONS
# ___________________________________


def run_main(api_key, model_key, model_parameters):
    task_id = call_start_api(api_key, model_key, model_parameters)
    # Just hardcode intervals
    while True:
        dict_out = call_check_api(api_key, task_id)
        if dict_out['data']['taskStatus'] == "Done":

            if "output" in dict_out['data']['taskOut']:
                return dict_out['data']['taskOut']["output"]
            else:
                return dict_out['data']['taskOut']
def start_main(api_key, model_key, model_parameters):
    task_id = call_start_api(api_key, model_key, model_parameters)
    return task_id
def check_main(api_key, task_id):
    dict_out = call_check_api(api_key, task_id)
    return dict_out


# THE API CALLING FUNCTIONS
# ________________________

# Takes in start params, returns task ID
def call_start_api(api_key, model_key, model_parameters):
    global endpoint
    route_start = "api/task/start/v1/"
    url_start = endpoint + route_start

    payload = {
        "id": str(uuid4()),
        "created": int(time.time()),
        "data": {
            "apiKey" : api_key,
            "modelKey" : model_key,
            "modelParameters" : model_parameters
        }
    }
    response = requests.post(url_start, json=payload)
    if response.status_code != 200:
        raise Exception("Server error: Booste inference server returned status code {}\n{}".format(
            response.status_code, response.json()['message']))
    try:
        out = response.json()
        if not out['success']:
            raise Exception(f"Booste inference start call failed with message: {out.message}")
        task_id = out['data']['taskID']
        return task_id
    except:
        raise Exception("Server error: Failed to return taskID")


# The bare async checker. Used by both gpt2_sync_main (automated) and async (client called)
# Takes in task ID, returns reformatted dict_out with Status and Output
def call_check_api(api_key, task_id):
    global endpoint
    route_check = "api/task/check/v1/"
    url_check = endpoint + route_check
    # Poll server for completed task

    payload = {
        "id": str(uuid4()),
        "created": int(time.time()),
        "longPoll": True,
        "data": {
            "taskID": task_id, 
            "apiKey": api_key
        }
    }
    response = requests.post(url_check, json=payload)
    if response.status_code != 200:
        raise Exception("Server error: Booste inference server returned status code {}\n{}".format(
            response.status_code, response.json()['message']))
    out = response.json()
    if not out['success']:
        raise Exception(f"Booste inference check call failed with message: {out.message}")
    return out


from distutils.core import setup
import setuptools
setup(
  name = 'booste',         # How you named your package folder (MyLib)
  packages = ['booste'],   # Chose the same as "name"
  version = '0.3.0',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'The booste package is a python client to interact with your machine learning models hosted on Booste',   # Give a short description about your library
  author = 'Erik Dunteman',                   # Type in your name
  author_email = 'erik@booste.io',      # Type in your E-Mail
  url = 'https://www.booste.io',   # Provide either the link to your github or to your website
  keywords = ['Booste client', 'API wrapper', 'Booste'],   # Keywords that define your package best
  setup_requires = ['wheel'],
  install_requires=[         # I get to this in a second
    'certifi',
    'chardet',
    'idna',
    'requests',
    'urllib3',
    'numpy',
    'pillow'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',

  ],
)
